# Prot386 Entry point and low-level functions
# (C) 2006-2007 Peter Ambroz
# This code is free software and is distributed under the terms of GNU GPL

	
	.text
	.globl _start
	.globl rm_reboot, reboot, sw_int
	.globl xdt_read, xdt_write, memcpy
	.globl load_ldtr, store_ldtr
	.globl getcpl
	.globl _gdt_base, _idt_base, _ldt_base, _tss_base
	.globl _utss_base, _supp_base

# Welcome to the new world. Everywere around, you can see vast ammounts of high memory.
# Setup your guards, populate the memory and live here in happiness until The Final Reset.

# This is were the protected mode code starts.
# bootloader did switch into PM by jumping here, now it's time to fire-up the application.

_start:
	movw $0x10, %ax
	movw %ax, %ds
	movw %ax, %es
	movw %ax, %fs
	movw %ax, %ss
	movl $0x7C00, %esp

	addw $8, %ax
	movw %ax, %gs	# Video-RAM segment
	
	cld
	
	# setup keyboard repeat rate to maximum
	subl $8, %esp
	movl $0x00, (%esp)
	movl $0xF3, 4(%esp)
	call kbd_send
	addl $8, %esp

	# hide the screen cursor (-> crt.s)
	call hide_cursor

	# setup interrupt handlers (-> idt.s)
	call idt_setup
	
	sti

	# setup TSS (Intel says it is necessary) (-> tss.s)
	# It will be necessary when it comes to privilege level changing calls
	call tss_setup

	# setup LDT (for experimenting with LDT) (-> tss.s)
	# Must go AFTER tss_setup
	call ldt_setup

	# setup supplementary User-space descriptors (-> tss.s)
	call extra_setup
	
	# initialize modules (-> modules.c)
	call init_modules

	# main application function (-> menu.c)
	call app_main

# Fast reboot code
reboot:
	movb $0xFE, %al
	outb %al, $0x64

# Halt, in case we didn't succeed
halt:	
	cli
	hlt
	jmp halt	# NMI can wake the processor from halt state

# Invoke software interrupt (dangerous, but handlers should withstand it)
# Works only when CS.Base == DS.Base
sw_int:
	movb	4(%esp), %ah	# parameter - int no.
	movb	$0xCD, %al	# opcode for int
	movw	%ax, _sw_int_jmp	# put the instruction in front of us (this awfuly breaks the prediction)
_sw_int_jmp:
	.word 0x0000
	ret

# void kbd_send(data, reg)
kbd_send:
	push %ebp
	movl %esp, %ebp
	
	call	empty_8042
	movb	12(%ebp), %al
	outb	%al, $0x60
	movb	8(%ebp), %al
	outb	%al, $0x60
	call	empty_8042
	
	leave
	ret
	
empty_8042:
	.word	0x00eb,0x00eb
	inb	$0x64, %al	# 8042 status port
	testb	$2, %al		# is input buffer full?
	jnz	empty_8042	# yes - loop
	ret

# read 8 byte descriptor from xDT[selector] into data
# void xdt_read(int selector, int table, int *data)
xdt_read:
	enter $0,$0
	push %edi
	push %esi
	movl 8(%ebp), %eax
	movl 12(%ebp), %edx
	movl 16(%ebp), %edi

	cmpl $0, %edx
	je _gdt_read
	cmpl $1, %edx
	je _ldt_read
	cmpl $2, %edx
	je _idt_read
_gdt_read:
	movl _gdt_base, %esi
	jmp _xdt_read_continue
_ldt_read:
	movl _ldt_base, %esi
	jmp _xdt_read_continue
_idt_read:
	movl _idt_base, %esi

_xdt_read_continue:
	shll $3, %eax
	addl %eax, %esi
	movsl
	movsl

	pop %esi
	pop %edi
	leave
	ret

# write 8 byte descriptor into xDT[selector]
# void xdt_write(int selector, int table, int *data)
xdt_write:
	enter $0,$0
	push %edi
	push %esi
	movl 8(%ebp), %eax
	movl 12(%ebp), %edx
	movl 16(%ebp), %esi

	cmpl $0, %edx
	je _gdt_write
	cmpl $1, %edx
	je _ldt_write
	cmpl $2, %edx
	je _idt_write
_gdt_write:
	movl _gdt_base, %edi
	jmp _xdt_write_continue
_ldt_write:
	movl _ldt_base, %edi
	jmp _xdt_write_continue
_idt_write:
	movl _idt_base, %edi

_xdt_write_continue:
	shll $3, %eax
	addl %eax, %edi
	movsl
	movsl

	pop %esi
	pop %edi
	leave
	ret

# Execute sldt and return result
load_ldtr:
	xor %eax, %eax
	sldt %ax
	shrl $3, %eax
	ret

# Execute lldt and update _ldtr_base pointer
# void store_ldtr(int selector)
store_ldtr:
	enter $0,$0

	movl 8(%ebp), %eax
	shll $3, %eax
	lldt %ax
	movl _tss_base, %edx
	movl %eax, 96(%edx)	# LDT field

	movl _gdt_base, %edx
	addl %eax, %edx
	movl (%edx), %eax
	movl 4(%edx), %ecx
	shrl $16, %eax
	andl $0xFF000000, %ecx
	addl %ecx, %eax
	movl %eax, _ldt_base

	leave
	ret

# get the current CPL
# int getcpl(void)
getcpl:
	movw %cs, %ax
	andl $0x03, %eax
	ret

# void *memcpy(void *dest, const void *src, int n)
memcpy:
	enter $0,$0
	push %esi
	push %edi

	movl 8(%ebp), %edi
	movl 12(%ebp), %esi
	movl 16(%ebp), %ecx
	movl %ecx, %edx

	andl $3, %edx
	shr $2, %ecx

	rep movsl
	movl %edx, %ecx
	rep movsb

	pop %edi
	pop %esi
	leave
	ret

.macro _memcpy dest src n
	subl $12, %esp
	movl \n, %eax
	movl %eax, 8(%esp)
	leal \src, %eax
	movl %eax, 4(%esp)
	leal \dest, %eax
	movl %eax, (%esp)
	call memcpy
	addl $12, %esp
.endm

# ACPI/APM powerdown with return to real-mode
# Return to real-mode. Say goodbye to high memory, goodbye descriptors, protection, privilege levels.. :(
rm_reboot:
	cli
 	_memcpy 0x7E00 rm_switch rm_switch_l
	_memcpy 0x8000 gdt_data $24
	lidt idt_48
	movw $0x10, %dx
	movw %dx, %ds
	movw %dx, %es
	movw %dx, %fs
	movw %dx, %gs
	movw %dx, %ss
	movl $0x7C00, %esp

	ljmp $0x8,$0x7E00

	hlt	
idt_48:
	.word	0x3FF	# idt limit=1024, appropriate for real-mode
	.long	0x0	# idt base

gdt_data:
	.long 0x00000000, 0x00000000	/* Null descriptor */
	.long 0x0000ffff, 0x00009a00	/* 16-bit real-mode 64k code */
	.long 0x0000ffff, 0x00009200	/* 16-bit real-mode 64k data */

# 16-bit real-mode switch and reboot
rm_switch:
.byte	0x66, 0x0f, 0x20, 0xc0		# movl  %cr0,%eax
.byte	0x66, 0x83, 0xe0, 0x00		# andl  $0x00000000,%eax
.byte	0x66, 0x0f, 0x22, 0xc0		# movl  %eax,%cr0
.byte	0xEA, 0x00, 0x00, 0xFF, 0xFF	# jmp 0xFFFF:0x0000 - BIOS reset code
rm_switch_l: .long . - rm_switch

_gdt_base:
	.long 0x8000
_idt_base:
	.long 0x8800
_ldt_base:
	.long 0x9000
_tss_base:
	.long 0x9800
_utss_base:
	.long 0xA000
_supp_base:
	.long 0xA800

